/* aliens.c setup, update and rendering of the aliens
 *
 * Copyright (c) 2019, Chris Camacho (codifies) http://bedroomcoders.co.uk/captcha/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "stdlib.h"

#include "raylib.h"
#include "raymath.h"

#include "aliens.h"
#include "screens.h"
#include "player.h"
#include "shots.h"
#include "game.h"

// aliens - setup and behaviour


static float formOffset;    // formation offset
static float formDir = -4.0;       // formation direction

//------------------------------------------------------------------------------
// update all the aliens
//------------------------------------------------------------------------------

void updateAliens()
{
    // a flag to see if we need a new level
    bool allDead = true;

    // formation offset update
    if (formOffset < (gameScreenWidth / 4.0)*3) {
        formDir = 4.0;
    }
    if (formOffset > (gameScreenWidth / 4.0)*5) {
        formDir = -4.0;
    }
    formOffset+=formDir;

    // "saucer"
    if (aliens[16].mode == am_Dead) {
        // if inactive 0.5% chance of new saucer (per frame)
        if (GetRandomValue(0,199)==0) {
            if (GetRandomValue(0,1)==0) {
                aliens[16].mode = am_SgoR;
                aliens[16].pos.x = -64;
                aliens[16].pos.y = 128;
            } else {
                aliens[16].mode = am_SgoL;
                aliens[16].pos.x = gameScreenWidth+64;
                aliens[16].pos.y = 128;
            }
            // only need one saucer sound at a time...
            PlaySound(saucerSound);
        }
    }


    for (int i=0; i<17; i++) {

        // saucer moving left
        if (aliens[i].mode == am_SgoL) {
            aliens[i].pos.x-=14;
            if (aliens[i].pos.x<-64) {
                aliens[i].mode = am_Dead;
            }
        }

        // saucer moving right
        if (aliens[i].mode == am_SgoR) {
            aliens[i].pos.x+=14;
            if (aliens[i].pos.x> gameScreenWidth+64) {
                aliens[i].mode = am_Dead;
            }
        }

        // if saucer moving theres a chance of it shooting
        if (aliens[i].mode == am_SgoR || aliens[i].mode == am_SgoL) {
            if (GetRandomValue(0,30)==0) {
                addShot(true, aliens[i].pos);
            }
        }

        // alien in formation behaviour
        if (aliens[i].mode == am_Form) {
            //aliens[i].rot = (atan2f(targ.y-aliens[i].pos.y, targ.x-aliens[i].pos.x)*(RAD2DEG))-90.0;
            aliens[i].rot = Vector2Angle(aliens[i].pos, player.pos) - 90.0;
            // formation is going in two different directions at the same time...
            if (i>7) {
                aliens[i].pos.x = formOffset-formationPos[i].x;
            } else {
                aliens[i].pos.x = (((gameScreenWidth/ 4.0)*8)-formOffset)-formationPos[i].x;
            }

            // should it dive out of formation
            if (player.mode == pm_Playing) {
                if (GetRandomValue(0,500)==0) {
                    aliens[i].mode = am_Dive;
                    aliens[i].diveTarget = player.pos.x;
                }
            }
        }

        // alien entering formation
        if (aliens[i].mode == am_Enter) {
            // rotation from alien to player
            aliens[i].rot = Vector2Angle(aliens[i].pos, player.pos) - 90.0;
            Vector2 ftarg; // formation target
            ftarg.y = formationPos[i].y;
            // formation is going in two different directions at the same time...
            if (i>7) {
                ftarg.x = formOffset-formationPos[i].x;
            } else {
                ftarg.x = (((gameScreenWidth/ 4.0)*8)-formOffset)-formationPos[i].x;
            }

            // move to the formation point or if there change mode
            Vector2 dir = Vector2Subtract(ftarg, aliens[i].pos);
            float l = Vector2Length(dir);
            if (l<8) {
                aliens[i].mode = am_Form;
            } else {
                dir = Vector2Normalize(dir);
                dir = Vector2Scale(dir, 6);
                aliens[i].pos = Vector2Add(aliens[i].pos, dir);
            }

        }

        // alien diving (rather similar to am_Enter....)
        if (aliens[i].mode == am_Dive) {
            if (player.mode != pm_Playing) {
                aliens[i].mode = am_Enter;
            }
            // rotation from alien to player
            aliens[i].rot = Vector2Angle(aliens[i].pos, player.pos) - 90.0;
            Vector2 ftarg; // dive target
            ftarg.x = aliens[i].diveTarget;
            ftarg.y = player.pos.y;

            // move to the dive point or if there change mode
            Vector2 dir = Vector2Subtract(ftarg, aliens[i].pos);
            float l = Vector2Length(dir);
            if (l<8) {
                aliens[i].mode = am_Enter;
            } else {
                dir = Vector2Normalize(dir);
                dir = Vector2Scale(dir, 6);
                aliens[i].pos = Vector2Add(aliens[i].pos, dir);
            }

        }


        // alien exploding sets to dead if explosion finished
        if (aliens[i].mode == am_Explode) {
            if (aliens[i].frame == 0) {
                PlaySoundEx(explosionSound);
                // if we had a pool of aliens a special alien could be created
                // (on a random chance) that was a bonus pickup
                // the player collides (not shoots) with a bonus
                // then they get some kind of benefit like temp shield or faster shooting
            }
            aliens[i].frame++;
            if (aliens[i].frame==25) {
                aliens[i].mode = am_Dead;
            }
        }

        // aliens shooting on a chance as long as they are not dead or exploding and the player is alive...
        if (GetRandomValue(1,75) == 1 &&
                aliens[i].mode != am_Dead &&
                aliens[i].mode != am_Explode &&
                player.mode != pm_Explode) {

            addShot(true, aliens[i].pos);
        }

        // any live aliens resets the all dead flag
        if(aliens[i].mode!=am_Dead) {
            allDead = false;
        }



        if (aliens[i].mode == am_Dive && player.mode == pm_Playing) {
            float d = Vector2Distance(aliens[i].pos, player.pos);
            if (d < 30) { // survive glancing blow ie not 32
                aliens[i].mode = am_Explode;
                aliens[i].frame = 0;
                player.mode = pm_Explode;
                player.frame = 0;
            }
        }

    } // end of for loop

    if (allDead) {
        // next wave... (could do something more interesting here...)
        // for example if we had a pool of aliens instead we could have
        // levels with even more aliens or very few really difficult aliens
        // by having an extra pathfinding challenge none shooting mode
        // challenge levels could be added
        resetPositions();
    }
}


//------------------------------------------------------------------------------
// reset all the aliens ready to come back on the screen
//------------------------------------------------------------------------------

void resetPositions()
{
    for (int i=0; i<16; i++) {

        if (GetRandomValue(0,1)==1) {
            aliens[i].pos.x = -400;
            aliens[i].pos.y = GetRandomValue(-gameScreenHeight,gameScreenHeight);
        } else {
            aliens[i].pos.x = gameScreenWidth + 400;
            aliens[i].pos.y = GetRandomValue(-gameScreenHeight,gameScreenHeight);
        }
        aliens[i].rot = 0;
        aliens[i].mode = am_Enter;
    }
    aliens[16].mode = am_Dead; // the "saucer"
}

void setupFormationPoints()
{
    initialiseShot();

    formOffset = (gameScreenWidth / 4.0) * 4;
    float v = (gameScreenWidth / 2.0) / 8.0;
    for (int i=0; i<8; i++) {
        formationPos[i] = (Vector2) { (gameScreenWidth / 4.0)+(v*i)+v/2.0, 300 };
        formationPos[i+8] = (Vector2) { (gameScreenWidth / 4.0)+(v*i)+v/2.0, 500 };
    }
}

void drawAliens()
{
    Rectangle src = (Rectangle){0,0,64,64};
    Rectangle dst = (Rectangle){0,0,64,64};
    Vector2 oset = (Vector2){32,32};

    for (int i=0; i<17; i++) {
        if (aliens[i].mode!=am_Dead) {
            dst.x = aliens[i].pos.x;
            dst.y = aliens[i].pos.y;
            if (aliens[i].mode!=am_Explode) {
                DrawTexturePro(aliens[i].tx, src, dst, oset, aliens[i].rot, WHITE);
            } else {
                // exploding
                src = (Rectangle){aliens[i].frame*64,0,64,64};
                DrawTexturePro(explosion, src, dst, oset, 0, WHITE);
                src = (Rectangle){0,0,64,64};
            }
        }
    }

}
