/* game.c loads and frees all the resources used in the actual game
 *
 * runs the game itself coordinating all the sub units  like player, aliens and shots
 *
 * Copyright (c) 2019, Chris Camacho (codifies) http://bedroomcoders.co.uk/captcha/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h> // for sprintf

#include "raylib.h"
#include "raymath.h"

#include "screens.h"
#include "aliens.h"
#include "player.h"
#include "shots.h"
#include "game.h"

// externed in screens.h
int gameScreenWidth = 1920;
int gameScreenHeight = 1080;

// its not worth a seperate unit for stars...
//
static const int numStars = 400;

// (deliberatly 2 white colours...)
Color starColours[8] = {
    { 230, 41, 55, 255 }, { 0, 228, 48, 255 }, { 0, 121, 241, 255 },
    {255, 255, 255, 255}, { 200, 122, 255, 255 }, { 253, 249, 0, 255 },
    { 255, 161, 0, 255 }, {255, 255, 255, 255}
};

// each star has its own position and colour
typedef struct star {
    Vector2 pos;
    Color colour;
} star;

// pointer to a block of memory for the stars
static star* stars;


static bool gameExit;           // does this screen want to exit
static ScreenCallbacks* scb;    // structure of callbacks for this screen


//------------------------------------------------------------------------------
// ask if this screen needs to exit
//------------------------------------------------------------------------------
static bool done()
{
    return gameExit;
}


// set the signal that the game screen should exit
// gameExit should only be modified here to ensure
// the cursor is reshown and the saucer sound is stopped
// also used by player unit
void setGameExit(bool b)
{
    gameExit = b;
    if (gameExit==true) {
        StopSound(saucerSound);
        ShowCursor();
    }
}

//------------------------------------------------------------------------------
// update stars, player, aliens, shots
//------------------------------------------------------------------------------
static void update()
{
    if (IsKeyDown(KEY_ESCAPE)) {
        setGameExit(true);
    }

    // update the stars position, if its left the screen
    // reposition and recolour it
    for (int i =0; i<numStars; i++) {
        stars[i].pos.y++;
        if (stars[i].pos.y > gameScreenHeight) {
            stars[i].pos.x = GetRandomValue(0, gameScreenWidth-1);
            stars[i].pos.y = 0;
            stars[i].colour = starColours[GetRandomValue(0,7)];
        }
    }

    updatePlayer();
    updateAliens();
    updateShots();

    checkShots(); // update shots could call this...

}

//------------------------------------------------------------------------------
// render
//------------------------------------------------------------------------------
static void render(fnt* font)
{
    ClearBackground(BLACK);

    // stuff rendered first is on the bottom
    // stuff rendered last is on the top!

    // render the starfield
    for (int i =0; i<numStars; i++) {
        // too small! DrawPixel(stars[i].pos.x, stars[i].pos.y, stars[i].colour);
        DrawRectangle(stars[i].pos.x, stars[i].pos.y, 4, 4, stars[i].colour);
    }

    drawShots();
    drawAliens();
    drawPlayer();  // also draws lives indicator

    // render text last...
    char str[1024];
    BeginShaderMode(font->fontShader);    // Activate SDF font shader
        // debug variables here...
        //sprintf(str,"last shot %i",player.lastShot);
        //DrawTextEx(font->font, str, (Vector2){2,600}, 20, -1, WHITE);

        sprintf(str,"SCORE: %i", score);
        DrawTextEx(font->font, str, (Vector2){200, gameScreenHeight-42}, 40, -1, WHITE);
        if (player.mode == pm_GameOver) {
            DrawTextEx(font->font, "GAME OVER",(Vector2){gameScreenWidth/2.0 - 400, gameScreenHeight/2.0-200}, 120, -1, WHITE);
            DrawTextEx(font->font, str, (Vector2){gameScreenWidth/2.0 - 400, gameScreenHeight/2.0-80}, 120, -1, WHITE);
        }
    EndShaderMode();

}


//------------------------------------------------------------------------------
// unload / free anything from this screens setup
//------------------------------------------------------------------------------
static void destroy()
{
    // free anything allocated in setup here
    RL_FREE(stars);

    UnloadTexture(alien1tx);
    UnloadTexture(alien2tx);
    UnloadTexture(playertx);

    // must stop the multichannel buffer pool before unloading
    // any sounds that might be playing in it...
    StopPlayBufferPool();

    UnloadSound(shotSound);
    UnloadSound(explosionSound);
    UnloadSound(saucerSound);

    destoryShotPool();

    free(scb);
}

//------------------------------------------------------------------------------
// reset this screens state
//------------------------------------------------------------------------------
static void reset()
{
    gameExit = false;

    // reset score etc here
    score = 0;

    // reset all aliens to start positions
    resetPositions();

    // reset the player
    player.mode = pm_Playing;
    player.lives = 3;
    player.pos.x = gameScreenWidth / 2.0;
    player.pos.y = gameScreenHeight -128;
    player.lastShot = 0;

    // delete all shots
    resetShots();

}


//------------------------------------------------------------------------------
// allocate and load stuff needed for this screen
//------------------------------------------------------------------------------

ScreenCallbacks* setUpGame()
{
    scb = malloc(sizeof(ScreenCallbacks));
    scb->update = update;
    scb->render = render;
    scb->destroy = destroy;
    scb->done = done;
    scb->reset = reset;

    gameExit = false;

    // remember to free anything allocated here in destroy() !
    // "unload" anything you "load"

    stars = RL_MALLOC(sizeof(star) * numStars);

    for (int i =0; i<numStars; i++) {
        star s;
        s.pos.x = GetRandomValue(0, gameScreenWidth-1);
        s.pos.y = GetRandomValue(0, gameScreenHeight-1);
        s.colour = starColours[GetRandomValue(0,7)];
        stars[i] = s;
    }

    alien1tx = LoadTexture("data/alien1.png");
    alien2tx = LoadTexture("data/alien2.png");

    playertx = LoadTexture("data/player.png");
    player.tx = playertx;

    aliens[16].tx = alien2tx;

    for (int i=0; i<16; i++) {
        aliens[i].tx = alien1tx;
    }

    setupFormationPoints();

    resetPositions();

    shotSound = LoadSound("data/shoot.wav");
    SetSoundVolume(shotSound, .5); // tone it down a bit (its loud)...

    saucerSound = LoadSound("data/saucer.wav");
    explosionSound = LoadSound("data/explosion.wav");

    explosion = LoadTexture("data/explosion.png");

    return scb;
}


