/* main.c the main "switch board" for the game it determines which "screen"
 * should be running calling their functions as apropriate
 *
 * Copyright (c) 2019, Chris Camacho (codifies) http://bedroomcoders.co.uk/captcha/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdlib.h>

#include "raylib.h"

#include "screens.h"


// settings "positions" in setting file
// TODO key binding etc here
typedef enum Settings { set_FULLSCREEN = 0 } Settings;

// font shared by the screens
fnt* mainFont;

int main(void)
{
    //--------------------------------------------------------------------------
    // setup
    //--------------------------------------------------------------------------

    // initial window size
    const int windowWidth = gameScreenWidth/2;
    const int windowHeight = gameScreenHeight/2;

    // Enable config flags for resizable window and vertical synchro
    SetConfigFlags(FLAG_WINDOW_RESIZABLE | FLAG_VSYNC_HINT | FLAG_MSAA_4X_HINT);
    InitWindow(windowWidth, windowHeight, "rayLib Shooter");
    SetWindowMinSize(320, 240);

    // get the fullscreen setting and change as apropriated
    fullScreen = StorageLoadValue(set_FULLSCREEN);
    if (fullScreen) {
        // set the window to the full resolution
        SetWindowSize(gameScreenWidth, gameScreenHeight);
        toggleScreen = 4; // wait before actually toggling
    } else {
        toggleScreen = 0;
    }

    SetExitKey(0);  // only the game screen allows explicit exit with escape
    InitAudioDevice();

    // Render texture initialization
    RenderTexture2D target = LoadRenderTexture(gameScreenWidth, gameScreenHeight);
    SetTextureFilter(target.texture, FILTER_BILINEAR);  // Texture scale filter to use

    // bundled shader and font together
    mainFont = calloc(1,sizeof(fnt));
    mainFont->font.baseSize = 16;
    mainFont->font.charsCount = 95;
    mainFont->font.chars = LoadFontData("data/04B_30__.TTF", 16, 0, 0, FONT_SDF);
    Image atlas = GenImageFontAtlas(mainFont->font.chars, 95, 32, 0, 1);
    mainFont->font.texture = LoadTextureFromImage(atlas);
    UnloadImage(atlas);
    mainFont->fontShader = LoadShader(0, "data/sdf.fs");
    SetTextureFilter(mainFont->font.texture, FILTER_BILINEAR);    // Required for SDF font

    SetTargetFPS(60);               // Set our game to run at 60 frames-per-second

    // setup the screen callbacks
    ScreenCallbacks* screenCalls[3];
    screenCalls[scr_SPLASH] = setUpSplash();
    screenCalls[scr_MENU] = setUpMenu();
    screenCalls[scr_GAME] = setUpGame();

    screenCalls[scr_GAME]->reset();

    currentScreen = scr_SPLASH;
    //currentScreen = scr_GAME;   // launch straight into game while developing

    //--------------------------------------------------------------------------
    // main loop
    //--------------------------------------------------------------------------

    TraceLog(LOG_INFO, "Starting main loop");
    // Main game loop

    while (!WindowShouldClose() && !screenCalls[scr_MENU]->done())
    {
        //----------------------------------------------------------------------
        // Update
        //----------------------------------------------------------------------

        // Compute required framebuffer scaling
        ScreenScale = (float)GetScreenWidth()/gameScreenWidth;
        float s = (float)GetScreenHeight()/gameScreenHeight;
        if ( s < ScreenScale) {
            ScreenScale = s;
        }

        // do the current screens update
        screenCalls[currentScreen]->update();

        //----------------------------------------------------------------------------------
        // Draw
        //----------------------------------------------------------------------------------
        BeginDrawing();

            ClearBackground(BLACK);

            // Draw everything in the render texture
            BeginTextureMode(target);
                // render the current screen
                screenCalls[currentScreen]->render(mainFont);
            EndTextureMode();

            // Draw RenderTexture2D to window, properly scaled and positioned
            DrawTexturePro(target.texture,
                    (Rectangle){
                        0.0f, 0.0f,
                        (float)target.texture.width, (float)-target.texture.height
                    },
                    (Rectangle){
                        (GetScreenWidth() - ((float)gameScreenWidth*ScreenScale))*0.5,
                        (GetScreenHeight() - ((float)gameScreenHeight*ScreenScale))*0.5,
                        (float)gameScreenWidth*ScreenScale,
                        (float)gameScreenHeight*ScreenScale
                    },
                    (Vector2){ 0, 0 }, 0.0f, WHITE);
        EndDrawing();

        // wait a few frames (or more) for the window resize to take effect
        // before toggling fullscreen
        if (toggleScreen!=0) {
            toggleScreen--;
            if (toggleScreen==0) {
                ToggleFullscreen();
                if (!fullScreen) {
                    // if not fullscreen resize and reposition the window
                    SetWindowSize(GetScreenWidth()/2, GetScreenHeight()/2);
                    SetWindowPosition(GetScreenWidth()/4, GetScreenHeight()/4);
                }
            }
        }

        // switch the sceen if we've finished with the current one
        if (screenCalls[currentScreen]->done()) {

            // exiting the menu screen quits (see start of main loop)
            if (currentScreen == scr_MENU) {
                screenCalls[scr_MENU]->destroy();
            }

            // exiting splash continues to the menu
            if (currentScreen == scr_SPLASH) {
                currentScreen = scr_MENU;
                screenCalls[scr_SPLASH]->destroy();
            }

            // exiting the game continues to the menu
            if (currentScreen == scr_GAME) {
                ShowCursor();
                StopPlayBufferPool();
                currentScreen = scr_MENU;
                screenCalls[scr_GAME]->reset();
            }

        }
    }
    //--------------------------------------------------------------------------------------
    // De-Initialization
    //--------------------------------------------------------------------------------------
    // we don't keep setting up and destroying the game screen each time so
    // destroy it here
    screenCalls[scr_GAME]->destroy();

    UnloadRenderTexture(target);    // Unload render texture
    UnloadShader(mainFont->fontShader);
    UnloadFont(mainFont->font); // unloading the font unloads its texture
    free(mainFont);

    CloseWindow();                  // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    // save settings
    StorageSaveValue(set_FULLSCREEN, fullScreen);

    return 0;
}
