/* shots.c setup, update and rendering of the shots
 *
 * Copyright (c) 2019, Chris Camacho (codifies) http://bedroomcoders.co.uk/captcha/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "raylib.h"
#include "raymath.h"

#include "shots.h"
#include "screens.h"
#include "player.h"
#include "aliens.h"
#include "game.h"

#include "PoolArray.h"

// this is the only unit that knows about "PoolArray" which is used to
// keep track of the the shot array


static shot shotArray[totalShots];
static PoolArray pa;

// deactivating all shots
void resetShots() {
    PoolArrayDeactivateAll(&pa);
}

// set up the PoolArray from the shotArray
void initialiseShot()
{
    PoolArrayInit(&pa, shotArray, sizeof(shot), totalShots );
}

void destoryShotPool()
{
    PoolArrayDestroy(pa);
}

// available shot list is checked to see if it has something available
// if so the shot is setup and added to the active list...
void addShot(bool isAlien, Vector2 p)
{
    // abort if not on screen yet
    if (p.x < 0 || p.x > gameScreenWidth) {
        return;
    }

    // is there an available shot it the pool
    if (isAlien) {
        if ( pa.Available - reservedShots < 1) {
            return;
        }
    } else {  // player shot
        if (!pa.Available) {
            return;
        }
    }

    // enlist a shot...
    shot* s = PoolArrayGetNextItem(&pa);
    if (!s) {
        // this will never happen as we are checking Available....
        return;
    }
    s->isAlien = isAlien;
    s->pos.x = p.x;
    s->pos.y = p.y;

    // use multi channel sound for shots...
    PlaySoundEx(shotSound);
}


// move the shots and retire them if they leave the screen
void updateShots()
{
    PoolArrayNode* node = pa.activeList;
    while(node) {
        PoolArrayNode* next = node->next;
        shot* s = (shot*)node->userData;
        if (s->isAlien) {
            s->pos.y += 8;
            if (s->pos.y>gameScreenHeight) {
                PoolArrayRetireNode(&pa, node);
            }
        } else {
            s->pos.y -= 8;
            if (s->pos.y<0) {
                PoolArrayRetireNode(&pa, node);
            }
        }
        node = next;
    }
}

// check for shot collisions for for aliens and player
void checkShots()
{
    PoolArrayNode* node = pa.activeList;
    while(node) {
        PoolArrayNode* next = node->next;
        shot* s = (shot*)node->userData;

        Vector2 sp;
        sp.x = s->pos.x;
        sp.y = s->pos.y;
        // is an alien shot too close to the player ?
        if (s->isAlien) {
            if (player.mode == pm_Playing) {
                float d = Vector2Distance(player.pos, sp);
                if ( d < 32.0) {
                    player.mode = pm_Explode;
                    player.frame = 0;
                    // retire the shot
                    PoolArrayRetireNode(&pa, node);
                }
            }
        } else { // is an alien too close to a players shot ?
            // check against *each* alien
            for (int i=0; i<17; i++) {
                if (aliens[i].mode!=am_Dead && aliens[i].mode!=am_Explode) {
                    // only check if its not dead and not exploding....
                    float d = Vector2Distance(aliens[i].pos, sp);
                    if (d < 32.0) {
                        score++;
                        if (i==16) {
                            // extra points for a saucer hit
                            score += 9;
                            StopSound(saucerSound);
                        }
                        aliens[i].mode = am_Explode;
                        aliens[i].frame = 0;
                        PoolArrayRetireNode(&pa, node);
                        // if two aliens were to die (near same position)
                        // we'd be trying to retire the shot twice, so let
                        // one live.... (break out of the alien loop)
                        // this was more important when the list used to free shots...
                        break;
                    }
                }
            }
        }
        node = next;
    }
}

// draw the shots on the active list
void drawShots()
{
    for(PoolArrayNode* node = pa.activeList; node; node = node->next) {
        shot* s = (shot*)node->userData;
        DrawCircle(s->pos.x, s->pos.y, 6, starColours[GetRandomValue(0,7)]);
    }
}
